"""Defines all functions related to Great Expectations Datasources."""
import click
import os
import yaml
import great_expectations as ge
from kedro_expectations.utils import base_ge_folder_does_NOT_exist, location_is_kedro_root_folder
from ..constants import _DEFAULT_PANDAS_DATASOURCE_NAME
from subprocess import Popen, DEVNULL


@click.command()
def init() -> None:
    if location_is_kedro_root_folder() and base_ge_folder_does_NOT_exist():
        init_ge_and_create_datasources()


def init_ge_and_create_datasources() -> None:
    try:
        print("Creating base great_expectations folder...")
        env = dict(os.environ)
        Popen(
            "echo Y | great_expectations init",
            shell=True,
            stdout=DEVNULL,
            env=env
        ).wait()
        print("great_expectations folder successfully created!")

        print("Generating Kedro Expectations Datasource...")

        context = ge.get_context()
        context.sources.add_pandas(name=_DEFAULT_PANDAS_DATASOURCE_NAME)

        print("Kedro Expectations successfully generated!")

    except yaml.YAMLError as exc:
        print("Error while parsing YAML:\n", exc)



import os
import glob

import pandas as pd
import yaml
import json
import great_expectations as ge
import datetime

from great_expectations.datasource.fluent import PandasDatasource
from kedro.framework.session import KedroSession
from .constants import _DEFAULT_PANDAS_DATASOURCE_NAME
import click
from time import sleep


def base_ge_folder_exists(verbose = True):
    base_folder = os.getcwd()
    ge_folder = os.path.join(base_folder, "great_expectations")
    if os.path.exists(ge_folder) is True:
        return True
    else:
        if verbose is True:
            message = """
            This command has NOT been run
            Kedro expectations wasn't initiated yet!
            Please run \'kedro expectations init\' before running this command.
            """
            print(message)
        return False


def base_ge_folder_does_NOT_exist(verbose = True):
    base_folder = os.getcwd()
    ge_folder = os.path.join(base_folder, "great_expectations")
    if os.path.exists(ge_folder) is False:
        return True
    else:
        if verbose == True:
            message = """
            This command has NOT been run
            Kedro expectations was already initiated and is ready to use.
            If you want to reset everything related to the plugin, you
            can delete the great_expectations folder and run init again
            """
            print(message)
        return False


def location_is_kedro_root_folder():
    try:
        project_path = os.getcwd()
        KedroSession.create(project_path=project_path)
        return True
    except ModuleNotFoundError:
        print("""
        Cannot run command!
        You need to be in a kedro root folder to use Kedro Expectations!
        """)
        return False


def is_dataset_in_catalog(input, catalog):
    if input in catalog.list():
        return True
    else:
        print(
            f"\n\nThe input {input} was not found at the DataCatalog.\n",
            "The following datasets are available for use:\n"
        )
        print(*catalog.list(), sep=', ')
        return False


def dot_to_underscore(value):
    adjusted_value = str(value).replace(".", "_")
    return adjusted_value


def get_property_from_catalog_item(catalog_item, item_property):
    current_dir_path = os.getcwd()
    catalog_path = os.path.join(
        current_dir_path,
        "conf",
        "base",
        "catalog.yml"
    )
    try:
        with open(catalog_path, "r") as stream:
            catalog = yaml.safe_load(stream)
            dataset = catalog[catalog_item]
            return dataset[item_property]
    except KeyError:
        pass


def get_or_add_pandas_datasource(ge_context, name=_DEFAULT_PANDAS_DATASOURCE_NAME):
    try:
        return ge_context.get_datasource(name)
    except ValueError:
        return ge_context.sources.add_pandas(name)


def get_or_add_dataframe_asset(datasource: PandasDatasource, name: str):
    try:
        return datasource.get_asset(name)
    except LookupError:
        return datasource.add_dataframe_asset(name)


def validate(adjusted_key, suite_name, validation_df):
    context = ge.get_context()
    formatted_time = datetime.datetime.now(
        datetime.timezone.utc
    ).strftime("%Y%m%dT%H%M%S")

    # Downward-compatibility: Make sure the dataframe asset exists
    dataframe_asset = get_or_add_dataframe_asset(get_or_add_pandas_datasource(context), name=adjusted_key)
    batch_request = dataframe_asset.build_batch_request(validation_df)

    checkpoint = context.add_or_update_checkpoint(
        name=f"{suite_name}_kedro_checkpoint",
        batch_request=batch_request,
        expectation_suite_name=suite_name
    )
    validation_result = checkpoint.run()

    return validation_result


def get_all_expectations(adjusted_key):
    exp_suites_path = os.path.join(
        os.getcwd(),
        "great_expectations",
        "expectations",
        adjusted_key,
    )
    all_expectations = glob.glob(
        os.path.join(
            exp_suites_path, "*.json"
        )
    )
    return all_expectations


def get_suite_name(exp_file, adjusted_key):
    parent_path, filename = os.path.split(exp_file)
    suite_name = adjusted_key + "." + filename[:-5]
    return suite_name


def create_raw_suite(adjusted_input, suite_name, expectation_suite_name):
    base_exp_path = os.path.join(
        os.getcwd(),
        "great_expectations",
        "expectations",
        adjusted_input,
    )

    if not os.path.exists(base_exp_path):
        os.makedirs(base_exp_path)

    new_suite_path = os.path.join(
        base_exp_path,
        suite_name+".json",
    )

    json_base = {
        "meta": {
            "great_expectations_version": "0.15.32",
        },
        "data_asset_type": None,
        "expectation_suite_name": expectation_suite_name,
        "expectations": [],
        "ge_cloud_id": None,
    }

    with open(new_suite_path, 'w') as f:
        json.dump(json_base, f, ensure_ascii=False, indent=4)


def populate_new_suite(input_data: pd.DataFrame, expectation_suite_name: str):
    ge_context = ge.data_context.DataContext()

    datasource = get_or_add_pandas_datasource(ge_context)
    data_asset = get_or_add_dataframe_asset(datasource, name=expectation_suite_name.split('.')[0])
    batch_request = data_asset.build_batch_request(dataframe=input_data)

    validator = ge_context.get_validator(
        batch_request=batch_request,
        expectation_suite_name=expectation_suite_name,
    )

    click.echo('\n\nYour dataset has the following columns:')
    click.echo(input_data.columns.values)
    click.echo('One by one, type the name of the columns you do NOT want to validate.\nOnce you are finished, '
               'type 0 to continue')
    column_to_remove = ""
    exclude_column_names = []
    while column_to_remove != "0":
        column_to_remove = click.prompt('', type=str)
        if column_to_remove == "0":
            pass
        elif column_to_remove not in input_data.columns:
            print(f"The column {column_to_remove} doesn't exist in this dataframe. Try typing again")
        else:
            exclude_column_names.append(column_to_remove)

    if exclude_column_names:
        print("The following columns are not going to be validated:")
        print(exclude_column_names)
        sleep(3)
    else:
        print("You chose for all columns to be validated!")
        sleep(3)

    # Removing duplicates
    exclude_column_names = [*set(exclude_column_names)]

    if len(exclude_column_names) >= len(input_data.columns.values):
        print(
            "\n\nAll the columns were marked to be excluded!",
            "Impossible to validate!"
        )
    else:
        result = ge_context.assistants.onboarding.run(
            batch_request=batch_request,
            exclude_column_names=exclude_column_names,
        )
        validator.expectation_suite = result.get_expectation_suite(
            expectation_suite_name=expectation_suite_name
        )
        validator.save_expectation_suite(discard_failed_expectations=False)
    print("\nFor more information about how to edit the expectations suite, access: "
          "https://docs.greatexpectations.io/docs/guides/expectations/creating_custom_expectations/overview/\n")


def choose_valid_suite_name():
    suite_name = "."
    while '.' in suite_name or ',' in suite_name or ' ' in suite_name:
        suite_name = click.prompt('', type=str)
        if '.' in suite_name or ' ' in suite_name:
            print(
                "Please choose another name for your suite.",
                "It cannot contain dots, commas or spaces"
            )
    return suite_name


def choose_valid_dataset_name():
    dataset_name = click.prompt('', type=str)
    is_partitioned = is_dataset_partitioned(dataset_name)
    while is_partitioned is not True:
        print(f"The dataset {dataset_name} is not partitioned! Type again:")
        dataset_name = click.prompt('', type=str)
        is_partitioned = is_dataset_partitioned(dataset_name)
    return dataset_name


def is_dataset_partitioned(dataset_name):
    if get_property_from_catalog_item(dataset_name, "type") == "PartitionedDataSet":
        return True
    else:
        return False



import click


@click.group(name="Kedro-Expectations")
def commands():
    """Kedro Expectations Command collection"""


@commands.group()
def expectations():
    """Run Kedro Expectations Commands"""


from .init_ge import init
from .create_suite import create_suite

expectations.add_command(init)
expectations.add_command(create_suite)

def main():
    commands()


if __name__ == "__main__":
    main()



from setuptools import find_packages, setup
from kedro_expectations import __version__

setup(
    name="kedro-expectations",
    version=__version__,
    url="https://gitlab.com/anacision-internal/kedro-great.git",
    author="Marcel Beining, based on work from Joao Gabriel Pampanin de Abreu",
    author_email="marcel.beining@anacision.de",
    description="Combine Kedro and Great Expectations",
    long_description="Combine Kedro and Great Expectations",
    long_description_content_type="text/markdown",
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    license="MIT",
    install_requires=[
        "kedro~=0.18",
        "great_expectations>=0.17.12",
        "pandas",
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    entry_points={
        "kedro.global_commands": ["kedro-expectations = kedro_expectations:commands"]
    }
)



import click
import os
from typing import Dict, cast, Callable
import pandas as pd
from kedro.framework.session import KedroSession
from kedro_expectations.utils import (dot_to_underscore, create_raw_suite, populate_new_suite,
                                      base_ge_folder_exists, location_is_kedro_root_folder, is_dataset_in_catalog,
                                      choose_valid_suite_name,
                                      get_property_from_catalog_item, choose_valid_dataset_name)


@click.command()
@click.option('--env', default=None, help='The respective Kedro environment to use')
def create_suite(env) -> None:
    # how to get cli args in here
    if location_is_kedro_root_folder() and base_ge_folder_exists():
        start_suite_creation(env=env)


def start_suite_creation(env=None):
    option = 0
    click.echo('Type 1 if you want to create a suite for a generic dataset')
    click.echo('Type 2 if you want to create a suite for a Partitioned dataset')
    option = click.prompt('', type=int)

    if option == 1:
        click.echo('Type the dataset name as it is on the DataCatalog')
        input = click.prompt('', type=str)

        click.echo('Type the desired name for the expectation suite')
        suite_name = choose_valid_suite_name()

        project_path = os.getcwd()
        with KedroSession.create(project_path=project_path,
                                 env=env,
                                 ) as session:
            kedro_context = session.load_context()
            catalog = kedro_context.catalog
            if is_dataset_in_catalog(input, catalog) is True:
                adjusted_input = dot_to_underscore(input)
                expectation_suite_name = adjusted_input + "." + suite_name

                create_raw_suite(adjusted_input, suite_name, expectation_suite_name)
                input_data = catalog.load(input)
                populate_new_suite(input_data, expectation_suite_name)

    elif option == 2:

        project_path = os.getcwd()
        with KedroSession.create(project_path=project_path,
                                 env=env
                                 ) as session:
            kedro_context = session.load_context()
            catalog = kedro_context.catalog

            partitioned_items = []
            for catalog_item in catalog.list():
                if get_property_from_catalog_item(catalog_item, "type") == "PartitionedDataSet":
                    partitioned_items.append(catalog_item)

        if partitioned_items:
            option = 0
            click.echo('\nType 1 if you want to create a generic expectation')
            click.echo('Type 2 if you want to create an specific expectation')
            option = click.prompt('', type=int)

            if option == 1:
                click.echo('Type the dataset name as it is on the DataCatalog')
                input = choose_valid_dataset_name()

                click.echo('Type the desired name for the expectation suite')
                suite_name = choose_valid_suite_name()

                project_path = os.getcwd()
                with KedroSession.create(project_path=project_path) as session:
                    kedro_context = session.load_context()
                    catalog = kedro_context.catalog
                    if is_dataset_in_catalog(input, catalog) is True:
                        adjusted_input = dot_to_underscore(input)
                        expectation_suite_name = adjusted_input + "." + suite_name

                        create_raw_suite(adjusted_input, suite_name, expectation_suite_name)
                        input_data = catalog.load(input)
                        partitions = cast(Dict[str, Callable], input_data)
                        validation_df = pd.DataFrame()

                        for casted_key, casted_value in partitions.items():
                            validation_df = pd.concat([validation_df, casted_value()], ignore_index=True, sort=False)
                            if len(validation_df.index) >= 5000:
                                break

                        populate_new_suite(validation_df, expectation_suite_name)
            elif option == 2:
                click.echo('Type the dataset name as it is on the DataCatalog')
                input = choose_valid_dataset_name()

                click.echo('Type the specific partition name you want to create an Expectation Suite for')
                desired_part = click.prompt('', type=str)

                click.echo('Type the desired name for the expectation suite')
                suite_name = choose_valid_suite_name()

                project_path = os.getcwd()
                with KedroSession.create(project_path=project_path) as session:
                    kedro_context = session.load_context()
                    catalog = kedro_context.catalog
                    if is_dataset_in_catalog(input, catalog) is True:
                        adjusted_input_pt1 = dot_to_underscore(input)
                        adjusted_input_pt2 = dot_to_underscore(desired_part)
                        expectation_suite_name = adjusted_input_pt1 + "." + adjusted_input_pt2 + "." + suite_name
                        adjusted_input = os.path.join(adjusted_input_pt1, adjusted_input_pt2)

                        create_raw_suite(adjusted_input, suite_name, expectation_suite_name)
                        input_data = catalog.load(input)
                        partitions = cast(Dict[str, Callable], input_data)

                        try:
                            validation_df = partitions[desired_part]()
                            print(validation_df.shape)
                            populate_new_suite(validation_df, expectation_suite_name)
                        except KeyError:
                            print(f"""
                                The partition {desired_part} does not exit!
                                Suite was not populated.
                                """)
            else:
                print("\n\nThe number typed is invalid. Aborting!")
        else:
            print("\n\nThere are no partitioned datasets registered in your catalog")
    else:
        print("\n\nThe number typed is invalid. Aborting!")



import os
from typing import Dict, Any, cast, Callable

from kedro.framework.hooks import hook_impl
from pandas import DataFrame as PandasDataFrame

from kedro_expectations.exceptions import SuiteValidationFailure
from kedro_expectations.utils import base_ge_folder_exists, location_is_kedro_root_folder, \
    get_property_from_catalog_item, dot_to_underscore, get_all_expectations, get_suite_name, validate


class KedroExpectationsHook:
    """Implementation of the Kedro Expectations Hooks."""

    def __init__(self, fail_fast: bool = False) -> None:
        self._fail_fast = fail_fast
        pass

    @hook_impl
    def before_node_run(self, inputs: Dict[str, Any]) -> None:
        """Validate inputs that are supported and have an expectation suite available."""
        if self.before_node_run and base_ge_folder_exists(verbose=False) and location_is_kedro_root_folder():
            self._run_validation(inputs)

    def _run_validation(self, data: Dict[str, Any]) -> None:
        for key, value in data.items():

            if get_property_from_catalog_item(key, "type") == "PartitionedDataSet":
                partitions = cast(Dict[str, Callable], value)

                for casted_key, casted_value in partitions.items():

                    # Looking for an specific expectation
                    adjusted_key_pt1 = dot_to_underscore(key)
                    adjusted_key_pt2 = dot_to_underscore(casted_key)

                    adjusted_key = os.path.join(adjusted_key_pt1, adjusted_key_pt2)
                    all_expectations = get_all_expectations(adjusted_key)

                    ge_adjusted_key = adjusted_key_pt1 + "." + adjusted_key_pt2

                    # Looking for a general expectation
                    if not all_expectations:
                        adjusted_key = dot_to_underscore(key)
                        all_expectations = get_all_expectations(adjusted_key)
                        ge_adjusted_key = adjusted_key

                    for exp_file in all_expectations:
                        suite_name = get_suite_name(exp_file, ge_adjusted_key)
                        result = validate(casted_key, suite_name, casted_value())

                        if self._fail_fast and not result.success:
                            raise SuiteValidationFailure(
                                f"Suite {suite_name} for DataSet {adjusted_key} failed!"
                            )
                    if not all_expectations:
                        print(
                            f"No expectation suite was found for \"{key}\".",
                            "Validation will be skipped!"
                        )
            else:
                adjusted_key = dot_to_underscore(key)
                all_expectations = get_all_expectations(adjusted_key)
                for exp_file in all_expectations:
                    suite_name = get_suite_name(exp_file, adjusted_key)

                    if isinstance(value, PandasDataFrame):
                        result = validate(adjusted_key, suite_name, value)
                    else:
                        print(f"Dataset {adjusted_key} is not supported by Kedro Expectations")

                    if self._fail_fast and not result.success:
                        raise SuiteValidationFailure(
                            f"Suite {suite_name} for DataSet {adjusted_key} failed!"
                        )
                if not all_expectations:
                    print(
                        f"No expectation suite was found for \"{key}\".",
                        "Validation will be skipped!"
                    )
